ITEM 'LevelBoundary'
CONCENTRATIONS 'Continuity' 'LevelBoundary'
DATA 1 1

ITEM 'Terminal'
CONCENTRATIONS 'Terminal' 'Continuity'
DATA 1 1

ITEM 'FlowBoundary'
CONCENTRATIONS 'Continuity' 'FlowBoundary'
DATA 1 1

ITEM 'Drainage'
CONCENTRATIONS 'Continuity' 'Drainage'
DATA 1 1

ITEM 'Precipitation'
CONCENTRATIONS 'Continuity' 'Precipitation'
DATA 1 1

ITEM 'UserDemand'
CONCENTRATIONS 'Continuity' 'UserDemand'
DATA 1 1

{% for boundary in boundaries -%}
ITEM '{{ boundary.name }}'
CONCENTRATIONS {{ boundary.substances | join(' ') | safe }}
ABSOLUTE TIME
LINEAR DATA {{ boundary.substances | join(' ') | safe }}
{{ boundary.df | safe }}

{% endfor -%}
{% for state in states -%}
ITEM '{{ state.name }}'
CONCENTRATIONS {{ state.substances | join(' ')}}
DATA {{ state.concentrations | join(' ')}}

{% endfor -%}
