ITEM 'LevelBoundary'
CONCENTRATIONS 'Continuity' 'LevelBoundary'
DATA 1 1

ITEM 'Terminal'
CONCENTRATIONS 'Terminal' 'Continuity'
DATA 1 1

ITEM 'FlowBoundary'
CONCENTRATIONS 'Continuity' 'FlowBoundary'
DATA 1 1

ITEM 'Drainage'
CONCENTRATIONS 'Continuity' 'Drainage'
DATA 1 1

ITEM 'Precipitation'
CONCENTRATIONS 'Continuity' 'Precipitation'
DATA 1 1

ITEM 'UserDemand'
CONCENTRATIONS 'Continuity' 'UserDemand'
DATA 1 1

ITEM 'LevelBoun_11'
CONCENTRATIONS 'Cl' 'Foo'
ABSOLUTE TIME
LINEAR DATA 'Cl' 'Foo'
'2020/01/01-00:00:00'  34.0  1.0

ITEM 'LevelBoun_17'
CONCENTRATIONS 'Cl'
ABSOLUTE TIME
LINEAR DATA 'Cl'
'2020/01/01-00:00:00'  34.0

ITEM 'FlowBound_15'
CONCENTRATIONS 'Bar' 'Cl' 'Tracer'
ABSOLUTE TIME
LINEAR DATA 'Bar' 'Cl' 'Tracer'
'2020/01/01-00:00:00'  1.0  0.0  1.0

ITEM 'FlowBound_16'
CONCENTRATIONS 'Cl' 'Tracer'
ABSOLUTE TIME
LINEAR DATA 'Cl' 'Tracer'
'2020/01/01-00:00:00'  0.0  1.0

ITEM 'Drainage_1'
CONCENTRATIONS 'Cl' 'Tracer'
ABSOLUTE TIME
LINEAR DATA 'Cl' 'Tracer'
'2020/01/01-00:00:00'  0.0    1.0
'2020/01/02-00:00:00'  1.0 -999.0

ITEM 'Precipita_1'
CONCENTRATIONS 'Cl' 'Tracer'
ABSOLUTE TIME
LINEAR DATA 'Cl' 'Tracer'
'2020/01/01-00:00:00'  0.0    1.0
'2020/01/02-00:00:00'  1.0 -999.0

ITEM 'Drainage_3'
CONCENTRATIONS 'Cl' 'Tracer'
ABSOLUTE TIME
LINEAR DATA 'Cl' 'Tracer'
'2020/01/01-00:00:00'  0.0    1.0
'2020/01/02-00:00:00'  1.0 -999.0

ITEM 'Precipita_3'
CONCENTRATIONS 'Cl' 'Tracer'
ABSOLUTE TIME
LINEAR DATA 'Cl' 'Tracer'
'2020/01/01-00:00:00'  0.0    1.0
'2020/01/02-00:00:00'  1.0 -999.0

ITEM 'Drainage_6'
CONCENTRATIONS 'Cl' 'Tracer'
ABSOLUTE TIME
LINEAR DATA 'Cl' 'Tracer'
'2020/01/01-00:00:00'  0.0    1.0
'2020/01/02-00:00:00'  1.0 -999.0

ITEM 'Precipita_6'
CONCENTRATIONS 'Cl' 'Tracer'
ABSOLUTE TIME
LINEAR DATA 'Cl' 'Tracer'
'2020/01/01-00:00:00'  0.0    1.0
'2020/01/02-00:00:00'  1.0 -999.0

ITEM 'Drainage_9'
CONCENTRATIONS 'Cl' 'Tracer'
ABSOLUTE TIME
LINEAR DATA 'Cl' 'Tracer'
'2020/01/01-00:00:00'  0.0    1.0
'2020/01/02-00:00:00'  1.0 -999.0

ITEM 'Precipita_9'
CONCENTRATIONS 'Cl' 'Tracer'
ABSOLUTE TIME
LINEAR DATA 'Cl' 'Tracer'
'2020/01/01-00:00:00'  0.0    1.0
'2020/01/02-00:00:00'  1.0 -999.0

