         ; DELWAQ_COMPLETE_ATTRIBUTES
    2    ; two blocks with input
    1    ; number of attributes, they are :
    1    ;  '1' is active '0' is not
    1    ; data follows in this file
    1    ; all data is given without defaults
   ;    layer:            1
{{ nsegments }}*1
    1    ; number of attributes, they are :
    2    ;  '1' has surface '3' has bottom
         ;  '0' has both    '2' has none
    1    ; data follows in this file
    1    ; all data is given without defaults
   ;    layer:            1
{{ nsegments }}*0
 0    ; no time dependent attributes
